# Event dataset
from typing import Iterable, Union
from urllib import parse
from urllib.parse import ParseResult

from .ObjectStream import ObjectStream
from .util_ast import as_ast, function_call


class EventDatasetURLException (BaseException):
    '''
    Exception thrown if the dataset URL passed is not valid
    '''
    def __init__(self, message):
        BaseException.__init__(self, message)


def _fixup_url(url: str, parsed_url) -> str:
    '''
    Fix up the url if we need to normalize anything

    Arguments:
        url         The URL to fix up
        parsed_url  The output of the URL parsing

    Returns:
        url         The URL fixed up. If it is a windows path, remove the extra
                    directory divider, for example.
    '''
    if parsed_url.scheme != 'file':
        return url

    # For file, we need to deal with file://path and file:///path.
    # If netloc is something we can quickly recognize as a local path or empty,
    # then this url is in good shape.
    if len(parsed_url.netloc) == 0 or parsed_url.netloc == 'localhost':
        return f'file://{parsed_url.path}'

    # Assume that netloc was part of the path.
    path = parsed_url.netloc
    if len(parsed_url.path) > 0:
        path = path + parsed_url.path
    return f'file:///{path}'


def _parse_and_check_dataset_url(u: str) -> ParseResult:
    '''
    Returns a parsed URL and the url, after checking the url's to make sure there
    is enough inforation in them. It throws if there is a problem.

    Args
        u       String

    Returns:
        u       The original URL
        parsed  The URL that was parsed by url2 lib.
    '''
    r = parse.urlparse(u)
    if r.scheme is None or len(r.scheme) == 0:
        raise EventDatasetURLException(f'EventDataSet({u}) has no scheme (file://, localds://, etc.)')
    if (r.netloc is None or len(r.netloc) == 0) and len(r.path) == 0:
        raise EventDatasetURLException(f'EventDataSet({u}) has no dataset or filename')

    return r


class EventDataset(ObjectStream):
    r'''
    Represents a stream of events that originates from a dataset specified by some sort of URL.
    '''
    def __init__(self, url: Union[str, Iterable[str]] = None):
        r'''
        Create and hold an event dataset reference. From one file, to multiple
        files, to a dataset specified otherwise.

        Args:
            url (str):  Must be a valid URL that points to a valid dataset

        Raises:
            Invalid URL
        '''
        if url is not None:
            # Normalize the URL as a list
            if isinstance(url, str):
                url = [url]
            l_url = list(url)

            if len(l_url) == 0:
                raise EventDatasetURLException("EventDataset initialized with an empty URL")

            # Make sure we can parse this URL. We don't, at some level, care about the actual contents.
            self.url = [_fixup_url(u, _parse_and_check_dataset_url(u)) for u in l_url]
        else:
            self.url = None

        # We participate in the AST parsing - as a node. So make sure the fields that should be traversed are
        # set.
        self._ast = function_call('EventDataset', [as_ast(self.url), ])



# Some ast utils
import ast
from typing import Any, List, Optional, Union, cast


def as_ast(p_var: Any) -> ast.AST:
    '''Convert any python constant into an ast

    Args:
        p_var   Some python variable that can be rendered with str(p_var)
                in a way the ast parse module will be able to ingest it.

    Returns:
        A python AST representing the object. For example, if a list is passed in, then
        the result will be an AST node of type ast.List.

    '''
    # If we are dealing with a string, we have to special case this.
    if isinstance(p_var, str):
        p_var = f"'{p_var}'"
    a = ast.parse(str(p_var))

    # Life out the thing inside the expression.
    # assert isinstance(a.body[0], ast.Expr)
    b = a.body[0]
    assert isinstance(b, ast.Expr)
    return b.value


def function_call(function_name: str, args: List[ast.AST]) -> ast.Call:
    '''
    Generate a function call to `function_name` with a list of `args`.

    Args:
        function_name   String that is the function name
        args            List of ast's, each one is an argument.
    '''
    return ast.Call(ast.Name(function_name, ast.Load()),
                    args,
                    [])


# TODO: lambda_unwrap should only be used in the parse_ast code, no where else - we should be moving
# Lambda AST's around, not Module AST's.
def lambda_unwrap(l: ast.AST) -> ast.Lambda:
    '''Given an AST of a lambda node, return the lambda node. If it is burried in a module, then unwrap it first
    Python, when it parses an module, returns the lambda wrapped in a `Module` AST node. This gets rid of it, but
    is also flexible.

    Args:
        l:      Lambda AST. It may be wrapped in a module as well.

    Returns:
        `Lambda` AST node, regardless of how the `Lambda` was wrapped in `l`.

    Exceptions:
        If the AST node isn't a lambda or a module wrapping a lambda.
    '''
    lb = cast(ast.Expr, l.body[0]).value if isinstance(l, ast.Module) else l
    if not isinstance(lb, ast.Lambda):
        raise BaseException('Attempt to get lambda expression body from {0}, which is not a lambda.'.format(type(l)))

    return lb


def lambda_args(l: Union[ast.Module, ast.Lambda]) -> ast.arguments:
    'Return the arguments of a lambda, no matter what form the lambda is in.'
    return lambda_unwrap(l).args


def lambda_body(l: Union[ast.Lambda, ast.Module]) -> ast.AST:
    '''
    Given an AST lambda node, get the expression it uses and return it. This just makes life easier,
    no real logic is occuring here.
    '''
    return lambda_unwrap(l).body


def lambda_call(args: Union[str, List[str]], l: Union[ast.Lambda, ast.Module]) -> ast.Call:
    '''
    Create a `Call` AST that calls a lambda with the named args.

    Args:
        args:       a single string or a list of strings, each string is an argument name to be passed in.
        l:          The lambda we want to call.

    Returns:
        A `Call` AST that calls the lambda with the given arguments.
    '''
    if isinstance(args, str):
        args = [args]
    named_args = [ast.Name(x, ast.Load()) for x in args]
    return ast.Call(lambda_unwrap(l), named_args, [])


def lambda_build(args: Union[str, List[str]], l_expr: ast.AST) -> ast.Lambda:
    '''
    Given a named argument(s), and an expression, build a `Lambda` AST node.

    Args:
        args:       the string names of the arguments to the lambda. May be a list or a single name
        l_expr:     An AST node that is the body of the lambda.

    Returns:
        The `Lambda` AST node.
    '''
    if type(args) is str:
        args = [args]

    ast_args = ast.arguments(args=[ast.arg(arg=x) for x in args])
    call_lambda = ast.Lambda(args=ast_args, body=l_expr)

    return call_lambda


def lambda_body_replace(l: ast.Lambda, new_expr: ast.AST) -> ast.Lambda:
    '''
    Return a new lambda function that has new_expr as the body rather than the old one. Otherwise, everything is the same.

    Args:
        l:          A ast.Lambda or ast.Module that points to a lambda.
        new_expr:   Expression that should replace this one.

    Returns:
        new_l: New lambda that looks just like the old one, other than the expression is new. If the old one was an ast.Module, so will this one be.
    '''
    if type(l) is not ast.Lambda:
        raise BaseException('Attempt to get lambda expression body from {0}, which is not a lambda.'.format(type(l)))

    new_l = ast.Lambda(l.args, new_expr)
    return new_l


def lambda_assure(east: ast.AST, nargs: Optional[int] = None):
    r'''
    Make sure the Python expression ast is a lambda call, and that it has the right number of args.

    Args:
        east:        python expression ast (module ast)
        nargs:      number of args it is required to have. If None, no check is done.
    '''
    if not lambda_test(east, nargs):
        raise BaseException(
            'Expression AST is not a lambda function with the right number of arguments')

    return east


def lambda_is_identity(l: ast.AST) -> bool:
    'Return true if this is a lambda with 1 argument that returns the argument'
    if not lambda_test(l, 1):
        return False

    b = lambda_unwrap(l)
    if not isinstance(b.body, ast.Name):
        return False

    a = b.args.args[0].arg
    return a == b.body.id


def lambda_is_true(l: ast.AST) -> bool:
    'Return true if this lambda always returns true'
    if not lambda_test(l):
        return False
    rl = lambda_unwrap(l)
    if not isinstance(rl.body, ast.NameConstant):
        return False

    return rl.body.value is True


def lambda_test(l: ast.AST, nargs: Optional[int] = None) -> bool:
    r''' Test arguments
    '''
    if not isinstance(l, ast.Lambda):
        if not isinstance(l, ast.Module):
            return False
        if len(l.body) != 1:
            return False
        if not isinstance(l.body[0], ast.Expr):
            return False
        if not isinstance(cast(ast.Expr, l.body[0]).value, ast.Lambda):
            return False
    rl = lambda_unwrap(l) if type(l) is ast.Module else l
    if type(rl) is not ast.Lambda:
        return False
    if nargs is None:
        return True
    return len(lambda_unwrap(l).args.args) == nargs



# Build a nice single cohesive version of the module with
# simple exports for the main items.

# flake8: noqa
# Because of how we pull things in, they are not used below, and so generate errors. We'll
# just turn off checking here.

# Main streaming object
from .ObjectStream import ObjectStream, ObjectStreamException  # NOQA

# Dataset accessors
from .EventDataset import EventDataset, EventDatasetURLException  # NOQA



# Helpers for LINQ operators and LINQ expressions in AST form.
# Utility routines to manipulate LINQ expressions.
import ast
from typing import Union

from .util_ast import lambda_unwrap


def parse_as_ast(ast_source: Union[str, ast.AST]) -> ast.Lambda:
    r'''Return an AST for a lambda function from several sources.

    We are handed one of several things:
        - An AST that is a lambda function
        - An AST that is a pointer to a Module that wraps an AST
        - Text that contains properly formatted ast code for a lambda function.

    In all cases, return a lambda function as an AST starting from the AST top node.

    Args:
        ast_source:     An AST or text string that represnets the lambda.

    Returns:
        An ast starting from the Lambda AST node.
    '''
    if isinstance(ast_source, str):
        a = ast.parse(ast_source.strip())
        return lambda_unwrap(a)
    else:
        return lambda_unwrap(ast_source)


# class ReplaceLINQOperators(ast.NodeTransformer):
#     r'''
#     A python 3 AST tranformer to replace function calls in the AST that are actually LINQ operators.

#     ObjectStream has methods called Select and SelectMany. When they are called, they build up the AST tree. But they do that
#     by creating Select and SelectMany, etc., ast nodes. When we parse a lambda passed as text, that does not happen. This
#     NodeTransformer does that replacement in-place.
#     '''

#     def visit_Call(self, node: ast.Call) -> Optional[ast.AST]:
#         '''Look for LINQ type calls and make a replacement with the appropriate AST entry
#         TODO: Make sure this is recursive properly!
#         '''
#         if isinstance(node.func, ast.Attribute):
#             func_name = node.func.attr
#             if func_name == "Select":
#                 source = self.visit(node.func.value)
#                 selection = self.visit(node.args[0])
#                 return Select(source, selection)
#             elif func_name == "SelectMany":
#                 source = self.visit(node.func.value)
#                 selection = self.visit(node.args[0])
#                 return SelectMany(source, selection)
#             elif func_name == "Where":
#                 source = self.visit(node.func.value)
#                 filter = self.visit(node.args[0])
#                 return Where(source, filter)
#             elif func_name == "First":
#                 source = self.visit(node.func.value)
#                 return First(source)
#             # Fall through to process the inside in the next step.
#         return self.generic_visit(node)



# An Object stream represents a stream of objects, floats, integers, etc.
import ast
import asyncio
from typing import Any, Callable, Union, cast

import nest_asyncio

from .util_ast import as_ast, function_call
from .util_ast_LINQ import parse_as_ast


class ObjectStreamException(BaseException):
    'Exception thrown by the ObjectStream object.'
    def __init__(self, msg):
        BaseException.__init__(self, msg)


class ObjectStream:
    r'''
    Represents the AST to produce a stream of objects. The objects can be events,
    jets, electrons, or just floats, or arrays of floats.

    `ObjectStream` holds onto the AST that will produce this stream of objects.
    '''
    def __init__(self, the_ast: ast.AST):
        r"""
        Initialize the stream with the ast that will produce this stream of objects.
        The user will almost never use this initializer.
        """
        self._ast = the_ast

    def SelectMany(self, func: Union[str, ast.Lambda]):
        r"""
        Given the current stream's object type is an array or other iterable, return
        the items in this objects type, one-by-one. This has the effect of flattening a
        nested array.

        Args:
            func:   The function that should be applied to this stream's objects to return
                    an iterable. Each item of the iterable is now the stream of objects.

        Returns:
            A new ObjectStream.
        """
        return ObjectStream(function_call("SelectMany", [self._ast, cast(ast.AST, parse_as_ast(func))]))

    def Select(self, f: Union[str, ast.Lambda]):
        r"""
        Apply a transformation function to each object in the stream, yielding a new type of
        object.

        Args:
            f:      selection function (lambda)

        Returns:
            A new ObjectStream of the transformed elements.
        """
        return ObjectStream(function_call("Select", [self._ast, cast(ast.AST, parse_as_ast(f))]))

    def Where(self, filter):
        r'''
        Filter the object stream, allowing only items for which `filter` evaluates as try through.

        Args:
            filter:     A filter lambda that returns True/False.

        Returns:
            A new ObjectStream that contains only elements that pass the filter function
        '''
        return ObjectStream(function_call("Where", [self._ast, cast(ast.AST, parse_as_ast(filter))]))

    def AsPandasDF(self, columns=[]):
        r"""
        Return a pandas dataframe. We do this by running the conversion.

        columns - Array of names of the columns. Will default to "col0", "call1", etc.
        """

        # To get Pandas use the ResultPandasDF function call.
        return ObjectStream(function_call("ResultPandasDF", [self._ast, as_ast(columns)]))

    def AsROOTTTree(self, filename, treename, columns=[]):
        r"""
        Return the sequence of items as a ROOT TTree. Each item in the ObjectStream
        will get one entry in the file. The items must be of types that the infrastructure
        can work with:
            Float:              A tree with a single float in each entry will be written.
            vector<float>:      A tree with a list of floats in each entry will be written.
            (<tuple>):          A tree with multiple items (leaves) will be written. Each leaf
                                must have one of the above types. Nested tuples are not supported.

        Args:
            filename:       Name of the file in which a TTree of the objects will be written.
            treename:       Name of the tree to be written to the file
            columns:        Array of names of the columns. This must match the number of items
                            in a tuple to be written out.

        Returns:
            A new ObjectStream with type [(filename, treename)]. This is because multiple tree's
            may be written by the back end, and need to be concatenated together to get the full
            dataset.
        """
        return ObjectStream(function_call("ResultTTree", [self._ast, as_ast(columns), as_ast(treename), as_ast(filename)]))

    def AsAwkwardArray(self, columns=[]):
        r'''
        Terminal - take the AST and return a root file.

        columns - Array of names of the columns
        '''
        return ObjectStream(function_call("ResultAwkwardArray", [self._ast, as_ast(columns)]))

    def _get_executor(self, executor: Callable[[ast.AST], Any] = None) -> Callable[[ast.AST], Any]:
        r'''
        Returns an executor that can be used to run this.
        Logic seperated out as it is used from several different places.

        Arguments:
            executor            Callback to run the AST. Can be synchronous or coroutine.

        Returns:
            An executor that is either synchronous or a coroutine.
        '''
        if executor is not None:
            return executor

        raise BaseException('No idea what to do for a default executor')

    async def _exe_as_task(self, executor: Callable[[ast.AST], Any]) -> Any:
        'Run the executor as a task, no matter if it is a co routine or not'
        r = executor(self._ast)

        if asyncio.iscoroutine(r):
            return await r
        else:
            return r

    async def value_async(self, executor: Callable[[ast.AST], Any] = None) -> None:
        r'''
        Start the evaluation of the AST. Returns a promise that can be used to check on the progress.
        Built to allow one to make lots of requests at the same time, and have a back-end server address
        them simultaneously.

        Args:
            executor:       A function that when called with the ast will return a future for the
                            result. If None, then uses the default executor (or throws if none is
                            defined).

        '''
        # Fetch the executor
        exe = self._get_executor(executor)

        # We do not know if this thing is synchronous or not, so we have to wrap it in a task.
        return await self._exe_as_task(exe)

    def value(self, executor: Callable[[ast.AST], Any] = None) -> Any:
        r"""
        Trigger the evaluation of the AST. Returns the results of the execution to the caller.
        WARNING: It is an error to call this if the event loop for async futures is already running.
                 This comes up most surprisingly when running in a Jupyter notebook, which starts an
                 event loop behind your back. So you must use `await` and `value_async`.

        Args:
            executor:       A function that when called with the ast will return the result. If
                            None, then use the default executor.

        Returns:
            Whatever the executor evaluates to.
        """
        # Use the nested asyncio package to get a running event loop.
        nest_asyncio.apply()
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.value_async(executor))
